@@ -225,8 +225,8 @@ def _get_neutron_port_from_docker_endpoint(endpoint_id):
 
 def _get_neutron_port_status_from_docker_endpoint(endpoint_id):
     response_port_status = {}
-    port_name = utils.get_neutron_port_name(endpoint_id)
-    filtered_ports = _get_ports_by_attrs(name=port_name)
+    neutron_port_identifier = _make_port_identifier(endpoint_id)
+    filtered_ports = _get_ports_by_identifier(neutron_port_identifier)
     if filtered_ports:
         response_port_status['status'] = filtered_ports[0]['status']
     return response_port_status
@@ -313,7 +313,9 @@ def _create_or_update_port(neutron_network_id, endpoint_id,
         port = filtered_ports['ports'][0]
         port_driver = get_driver(port)
         response_port = port_driver.update_port(port, endpoint_id,
-                                                interface_mac)
+                                                interface_mac,
+                                                tags=app.tag)
+        _neutron_port_add_tags(port, endpoint_id)
     # For the container boot from dual-net, request_address will
     # create two ports(v4 and v6 address), we should only allow one
     # for port bind.
@@ -333,7 +335,9 @@ def _create_or_update_port(neutron_network_id, endpoint_id,
             else:
                 port_driver = get_driver(port)
                 response_port = port_driver.update_port(port, endpoint_id,
-                                                        interface_mac)
+                                                        interface_mac,
+                                                        tags=app.tag)
+                _neutron_port_add_tags(port, endpoint_id)
         if not response_port:
             fixed_ips = (
                 lib_utils.get_dict_format_fixed_ips_from_kv_format(fixed_ips))
@@ -347,62 +351,78 @@ def _create_or_update_port(neutron_network_id, endpoint_id,
     return response_port, subnets
 
 
-def _neutron_net_add_tag(netid, tag):
-    _neutron_add_tag('networks', netid, tag)
+def _neutron_net_add_tag(net, tag):
+    _neutron_add_tag('networks', net, tag)
 
 
-def _neutron_net_add_tags(netid, tag, tags=True):
+def _neutron_net_add_tags(net, tag, tags=True):
     if tags:
         tags = utils.create_net_tags(tag)
         for tag in tags:
-            _neutron_net_add_tag(netid, tag)
+            _neutron_net_add_tag(net, tag)
 
 
-def _neutron_net_remove_tag(netid, tag):
-    _neutron_remove_tag('networks', netid, tag)
+def _neutron_net_remove_tag(net, tag):
+    _neutron_remove_tag('networks', net, tag)
 
 
-def _neutron_net_remove_tags(netid, tag):
+def _neutron_net_remove_tags(net, tag):
     tags = utils.create_net_tags(tag)
     for tag in tags:
-        _neutron_net_remove_tag(netid, tag)
+        _neutron_net_remove_tag(net, tag)
 
 
-def _neutron_subnetpool_add_tag(poolid, tag):
-    _neutron_add_tag('subnetpools', poolid, tag)
+def _neutron_subnetpool_add_tag(pool, tag):
+    _neutron_add_tag('subnetpools', pool, tag)
 
 
-def _neutron_subnetpool_remove_tag(poolid, tag):
-    _neutron_remove_tag('subnetpools', poolid, tag)
+def _neutron_subnetpool_remove_tag(pool, tag):
+    _neutron_remove_tag('subnetpools', pool, tag)
 
 
-def _neutron_subnet_add_tag(subnetid, tag):
-    _neutron_add_tag('subnets', subnetid, tag)
+def _neutron_subnet_add_tag(subnet, tag):
+    _neutron_add_tag('subnets', subnet, tag)
 
 
-def _neutron_subnet_remove_tag(subnetid, tag):
-    _neutron_remove_tag('subnets', subnetid, tag)
+def _neutron_subnet_remove_tag(subnet, tag):
+    _neutron_remove_tag('subnets', subnet, tag)
 
 
-def _neutron_port_add_tag(portid, tag):
-    _neutron_add_tag('ports', portid, tag)
+def _neutron_port_add_tag(port, tag):
+    _neutron_add_tag('ports', port, tag)
 
 
-def _neutron_port_remove_tag(portid, tag):
-    _neutron_remove_tag('ports', portid, tag)
+def _neutron_port_add_tags(port, tag):
+    if app.tag:
+        tags = utils.create_port_tags(tag)
+        for tag in tags:
+            _neutron_port_add_tag(port, tag)
 
 
-def _neutron_add_tag(resource_type, resource_id, tag):
-    try:
-        app.neutron.add_tag(resource_type, resource_id, tag)
-    except n_exceptions.NotFound:
-        LOG.warning("Neutron tags extension for given "
-                    "resource type is not supported, "
-                    "cannot add tag to %s.", resource_type)
+def _neutron_port_remove_tag(port, tag):
+    _neutron_remove_tag('ports', port, tag)
 
 
-def _neutron_remove_tag(resource_type, resource_id, tag):
-    app.neutron.remove_tag(resource_type, resource_id, tag)
+def _neutron_port_remove_tags(port, tag):
+    if app.tag:
+        tags = utils.create_port_tags(tag)
+        for tag in tags:
+            _neutron_port_remove_tag(port, tag)
+
+
+def _neutron_add_tag(resource_type, resource, tag):
+    if tag not in resource['tags']:
+        try:
+            app.neutron.add_tag(resource_type, resource['id'], tag)
+        except n_exceptions.NotFound:
+            LOG.warning("Neutron tags extension for given "
+                        "resource type is not supported, "
+                        "cannot add tag to %s.", resource_type)
+
+
+def _neutron_remove_tag(resource_type, resource, tag):
+    if tag in resource['tags']:
+        app.neutron.remove_tag(resource_type, resource['id'], tag)
 
 
 def _make_net_identifier(network_id, tags=True):
@@ -417,6 +437,18 @@ def _get_networks_by_identifier(identifier):
     return _get_networks_by_attrs(name=identifier)
 
 
+def _make_port_identifier(endpoint_id):
+    if app.tag:
+        return utils.make_port_tags(endpoint_id)
+    return utils.get_neutron_port_name(endpoint_id)
+
+
+def _get_ports_by_identifier(identifier):
+    if app.tag:
+        return _get_ports_by_attrs(tags=identifier)
+    return _get_ports_by_attrs(name=identifier)
+
+
 def _program_expose_ports(options, port_id):
     exposed_ports = options.get(const.DOCKER_EXPOSED_PORTS_OPTION)
     if not exposed_ports:
@@ -517,33 +549,6 @@ def _get_cidr_from_subnetpool(**kwargs):
             .format(kwargs))
 
 
-def _update_existing_port(existing_port, fixed_ip, mac_address):
-    host = existing_port.get('binding:host_id')
-    vif_type = existing_port.get('binding:vif_type')
-    if not host and vif_type == 'unbound':
-        updated_port = {
-            'name': const.NEUTRON_UNBOUND_PORT,
-            'admin_state_up': True,
-        }
-        if mac_address:
-            updated_port['mac_address'] = mac_address
-        updated_port_resp = app.neutron.update_port(
-            existing_port['id'],
-            {'port': updated_port})
-        existing_port = updated_port_resp['port']
-    else:
-        port_name = existing_port.get('name', '')
-        if (str(port_name) != const.NEUTRON_UNBOUND_PORT or
-                len(existing_port['fixed_ips']) <= 1 or
-                host != lib_utils.get_hostname()):
-            raise exceptions.AddressInUseException(
-                "Requested ip address {0} already belongs to "
-                "a bound Neutron port: {1}".format(fixed_ip,
-                                                   existing_port['id']))
-
-    return existing_port
-
-
 def revoke_expose_ports(port_id):
     sgs = app.neutron.list_security_groups(
         name=utils.get_sg_expose_name(port_id))
@@ -591,7 +596,7 @@ def _create_kuryr_subnet(pool_cidr, subnet_cidr, pool_id, network_id, gateway):
     subnets = app.neutron.create_subnet({'subnets': new_kuryr_subnet})
     if app.tag_ext:
         for subnet in subnets['subnets']:
-            _neutron_subnet_add_tag(subnet['id'], pool_id)
+            _neutron_subnet_add_tag(subnet, pool_id)
     LOG.debug("Created kuryr subnet %s", new_kuryr_subnet)
 
 
@@ -619,7 +624,7 @@ def _create_kuryr_subnetpool(pool_cidr, pool_tag, shared):
         {'subnetpool': new_subnetpool})
     pool = created_subnetpool_response['subnetpool']
     if pool_tag:
-        _neutron_subnetpool_add_tag(pool['id'], pool_tag)
+        _neutron_subnetpool_add_tag(pool, pool_tag)
     return pool
 
 
@@ -881,9 +886,9 @@ def network_driver_create_network():
         network = app.neutron.create_network(
             {'network': {'name': neutron_network_name,
                          "admin_state_up": True,
-                         'shared': shared}})
-        network_id = network['network']['id']
-        _neutron_net_add_tags(network['network']['id'], container_net_id,
+                         'shared': shared}})['network']
+        network_id = network['id']
+        _neutron_net_add_tags(network, container_net_id,
                               tags=app.tag)
 
         LOG.info("Created a new network with name "
@@ -902,15 +907,16 @@ def network_driver_create_network():
                 raise exceptions.KuryrException(
                     ("Specified network id/name({0}) does not "
                      "exist.").format(specified_network))
-            network_id = networks[0]['id']
+            network = networks[0]
+            network_id = network['id']
             network_shared = networks[0]['shared']
         except n_exceptions.NeutronClientException as ex:
             LOG.error("Error happened during listing "
                       "Neutron networks: %s", ex)
             raise
         if app.tag:
-            _neutron_net_add_tags(network_id, container_net_id, tags=app.tag)
-            _neutron_net_add_tag(network_id,
+            _neutron_net_add_tags(network, container_net_id, tags=app.tag)
+            _neutron_net_add_tag(network,
                                  utils.existing_net_tag(container_net_id))
         else:
             network = app.neutron.update_network(
@@ -951,7 +957,7 @@ def network_driver_create_network():
 
     def _add_tag_for_existing_subnet(subnet, pool_id):
         if len(subnet) == 1:
-            _neutron_subnet_add_tag(subnet[0]['id'], pool_id)
+            _neutron_subnet_add_tag(subnet[0], pool_id)
 
     # This will add a subnetpool_id(created by kuryr) tag
     # for existing Neutron subnets.
@@ -1007,9 +1013,10 @@ def network_driver_delete_network():
             LOG.warning("Network is a pre existing Neutron "
                         "network, not deleting in Neutron. "
                         "removing tags: %s", existing_network_identifier)
-            neutron_net_id = existing_networks[0]['id']
-            _neutron_net_remove_tags(neutron_net_id, container_net_id)
-            _neutron_net_remove_tag(neutron_net_id,
+            neutron_net = existing_networks[0]
+            neutron_net_id = neutron_net['id']
+            _neutron_net_remove_tags(neutron_net, container_net_id)
+            _neutron_net_remove_tag(neutron_net,
                                     utils.existing_net_tag(container_net_id))
             # Delete subnets created by kuryr
             filtered_subnets = _get_subnets_by_attrs(
@@ -1260,12 +1267,12 @@ def network_driver_delete_endpoint():
             .format(neutron_network_identifier)
         })
     else:
-        neutron_port_name = utils.get_neutron_port_name(endpoint_id)
-        filtered_ports = _get_ports_by_attrs(name=neutron_port_name)
+        neutron_port_identifier = _make_port_identifier(endpoint_id)
+        filtered_ports = _get_ports_by_identifier(neutron_port_identifier)
         if not filtered_ports:
             raise exceptions.NoResourceException(
-                "The port doesn't exist for the name {0}"
-                .format(neutron_port_name))
+                "The port doesn't exist for the identifier {0}"
+                .format(neutron_port_identifier))
         neutron_port = filtered_ports[0]
 
         try:
@@ -1287,6 +1294,8 @@ def network_driver_delete_endpoint():
             with excutils.save_and_reraise_exception():
                 LOG.error('Error while removing the interface: %s', ex)
 
+        _neutron_port_remove_tags(neutron_port, endpoint_id)
+
     return flask.jsonify(const.SCHEMA['SUCCESS'])
 
 
@@ -1344,12 +1353,12 @@ def network_driver_join():
     else:
         neutron_network_id = filtered_networks[0]['id']
 
-        neutron_port_name = utils.get_neutron_port_name(endpoint_id)
-        filtered_ports = _get_ports_by_attrs(name=neutron_port_name)
+        neutron_port_identifier = _make_port_identifier(endpoint_id)
+        filtered_ports = _get_ports_by_identifier(neutron_port_identifier)
         if not filtered_ports:
             raise exceptions.NoResourceException(
-                "The port doesn't exist for the name {0}"
-                .format(neutron_port_name))
+                "The port doesn't exist for the identifier {0}"
+                .format(neutron_port_identifier))
         neutron_port = filtered_ports[0]
         all_subnets = _get_subnets_by_attrs(network_id=neutron_network_id)
         kuryr_subnets = []
@@ -1585,7 +1594,7 @@ def ipam_request_pool():
             pool_id = existing_pools[0]['id']
             if app.tag_ext:
                 _neutron_subnetpool_add_tag(
-                    pool_id, const.KURYR_EXISTING_NEUTRON_SUBNETPOOL)
+                    existing_pools[0], const.KURYR_EXISTING_NEUTRON_SUBNETPOOL)
             prefixes = existing_pools[0]['prefixes']
             pool_cidr = ipaddress.ip_network(six.text_type(prefixes[0]))
             if pool_cidr == cidr:
@@ -1735,21 +1744,19 @@ def ipam_request_address():
                 fixed_ips.append(fixed_ip)
 
                 if num_ports:
-                    existing_port = filtered_ports[0]
-                    created_port = _update_existing_port(
-                        existing_port, fixed_ip, req_mac_address)
+                    created_port = filtered_ports[0]
                     # REVISIT(yedongcan) For tag-ext extension not
                     # supported, the Neutron existing port still can not
                     # be deleted in ipam_release_address.
                     if app.tag_ext:
                         _neutron_port_add_tag(
-                            created_port['id'],
+                            created_port,
                             const.KURYR_EXISTING_NEUTRON_PORT)
                 else:
                     created_port_resp = app.neutron.create_port({'port': port})
                     created_port = created_port_resp['port']
                     if app.tag_ext:
-                        _neutron_port_add_tag(created_port['id'],
+                        _neutron_port_add_tag(created_port,
                                               lib_const.DEVICE_OWNER)
 
                 LOG.debug("created port %s", created_port)
@@ -1810,7 +1817,7 @@ def ipam_release_pool():
             cidr=six.text_type(subnet_cidr))
         for tmp_subnet in subnets_by_cidr:
             if pool_id in tmp_subnet.get('tags', []):
-                _neutron_subnet_remove_tag(tmp_subnet['id'], pool_id)
+                _neutron_subnet_remove_tag(tmp_subnet, pool_id)
                 break
 
     pools = _get_subnetpools_by_attrs(id=pool_id)
@@ -1820,7 +1827,7 @@ def ipam_release_pool():
             tags = pools[0].get('tags', [])
             if const.KURYR_EXISTING_NEUTRON_SUBNETPOOL in tags:
                 _neutron_subnetpool_remove_tag(
-                    pool_id, const.KURYR_EXISTING_NEUTRON_SUBNETPOOL)
+                    pools[0], const.KURYR_EXISTING_NEUTRON_SUBNETPOOL)
                 LOG.debug('Skip the cleanup since this is an existing Neutron '
                           'subnetpool.')
                 return flask.jsonify(const.SCHEMA['SUCCESS'])
@@ -1896,7 +1903,7 @@ def ipam_release_address():
                     if (port['fixed_ips'][0]['subnet_id'] == tmp_subnet['id']):
                         app.neutron.delete_port(port['id'])
             elif tags and const.KURYR_EXISTING_NEUTRON_PORT in tags:
-                updated_port = {'name': '', 'device_owner': '',
+                updated_port = {'device_owner': '',
                                 'binding:host_id': ''}
                 if port['name'].startswith(port['device_id']):
                     updated_port["device_id"] = ''
@@ -1905,7 +1912,7 @@ def ipam_release_address():
                         app.neutron.update_port(
                             port['id'], {'port': updated_port})
                         _neutron_port_remove_tag(
-                            port['id'], const.KURYR_EXISTING_NEUTRON_PORT)
+                            port, const.KURYR_EXISTING_NEUTRON_PORT)
     except n_exceptions.NeutronClientException as ex:
         LOG.error("Error happened while fetching "
                   "and deleting port, %s", ex)
