@@ -99,10 +99,10 @@ class Driver(object):
         raise NotImplementedError()
 
     @abc.abstractmethod
-    def get_container_iface_name(self, neutron_port_id):
+    def get_container_iface_name(self, neutron_port):
         """Returns interface name of a container in the default namespace.
 
-        :param neutron_port_id: The ID of a neutron port as string
+        :param neutron_port: The neutron port
         :returns: interface name as string
         """
         raise NotImplementedError()
@@ -115,7 +115,7 @@ class Driver(object):
         """
         raise NotImplementedError()
 
-    def update_port(self, port, endpoint_id, interface_mac):
+    def update_port(self, port, endpoint_id, interface_mac, tags=True):
         """Updates port information and performs extra driver-specific actions.
 
         It returns the updated port dictionary after the required actions
@@ -128,35 +128,49 @@ class Driver(object):
         :returns: the updated Neutron port id dictionary as returned by
                   python-neutronclient
         """
-        port['name'] = libnet_utils.get_neutron_port_name(endpoint_id)
         try:
-            updated_port = {
-                'name': port['name'],
-                'device_owner': lib_const.DEVICE_OWNER,
-                'device_id': endpoint_id,
-                'binding:host_id': lib_utils.get_hostname(),
-            }
-            if interface_mac:
+            updated_port = {}
+            hostname = lib_utils.get_hostname()
+            if port['binding:host_id'] != hostname:
+                updated_port['binding:host_id'] = hostname
+                updated_port['device_owner'] = lib_const.DEVICE_OWNER
+            if port['admin_state_up'] is not True:
+                updated_port['admin_state_up'] = True
+            if not tags:
+                # rename the port if tagging is not supported
+                updated_port['name'] = libnet_utils.get_neutron_port_name(
+                    endpoint_id)
+            if not port.get('device_id'):
+                updated_port['device_id'] = endpoint_id
+            if interface_mac and port['mac_address'] != interface_mac:
                 updated_port['mac_address'] = interface_mac
-            response_port = app.neutron.update_port(port['id'],
-                                                    {'port': updated_port})
+            if updated_port:
+                port = app.neutron.update_port(port['id'],
+                                               {'port': updated_port})['port']
         except n_exceptions.NeutronClientException as ex:
             LOG.error("Error happened during updating a "
                       "Neutron port: %s", ex)
             raise
-        return response_port['port']
+        return port
 
     def __str__(self):
         return self.__class__.__name__
 
 
-def get_driver_instance():
+def get_driver_instance(name=None):
     """Instantiate a driver instance accordingly to the file configuration.
 
     :returns: a Driver instance
     :raises: exceptions.KuryrException
     """
-    module, name, classname = _parse_port_driver_config()
+    if name:
+        module, name, classname = _parse_port_driver_config(name)
+    else:
+        module, name, classname = _parse_port_driver_config()
+
+    if (module not in config.CONF.enabled_port_drivers and
+            name not in config.CONF.enabled_port_drivers):
+        raise exceptions.KuryrException("No port driver available")
 
     # TODO(apuimedo): switch to the openstack/stevedore plugin system
     try:
@@ -171,13 +185,14 @@ def get_driver_instance():
     return driver
 
 
-def _parse_port_driver_config():
+def _parse_port_driver_config(config_value=None):
     """Read the port driver related config value and parse it.
 
     :returns: the provided full module path as per config file, the name of the
              driver/module and the class name of the Driver class inside it
     """
-    config_value = config.CONF.port_driver
+    if config_value is None:
+        config_value = config.CONF.default_port_driver
     config_tokens = config_value.rsplit('.', 1)
     if len(config_tokens) == 1:  # not a path, just a name
         name = config_tokens[0]
@@ -203,21 +218,23 @@ def _verify_port_driver_compliancy(driver, port_driver_name):
 
 
 def _verify_binding_driver_compatibility(driver, port_driver_name):
-    tokens = config.CONF.binding.driver.rsplit('.', 1)
-    binding_driver_name = tokens[0] if len(tokens) == 1 else tokens[1]
-    binding_driver_name.lower()
+    binding_drivers_names = []
+    for binding_driver in config.CONF.binding.enabled_drivers:
+        tokens = binding_driver.rsplit('.', 1)
+        binding_driver_name = tokens[0] if len(tokens) == 1 else tokens[1]
+        binding_drivers_names.append(binding_driver_name.lower())
 
     # TODO(mchiappe): find a clean way to test the binding driver
     # is also loadable before we start
     supported_bindings = driver.get_supported_bindings()
 
-    if binding_driver_name not in supported_bindings:
+    if not set(binding_drivers_names) & set(supported_bindings):
         raise exceptions.KuryrException("Configuration file error: "
             "port driver '{0}' is not compatible with binding driver '{1}'"
-            .format(port_driver_name, binding_driver_name))
+            .format(port_driver_name, binding_drivers_names))
 
     # Temporarily ban IPVLAN, to be removed in the future
-    if binding_driver_name == 'ipvlan':
+    if 'ipvlan' in binding_drivers_names:
         raise exceptions.KuryrException("Configuration file error: "
             "binding driver '{0}' is currently not supported with '{1}' "
-            "port driver".format(binding_driver_name, port_driver_name))
+            "port driver".format(binding_drivers_names, port_driver_name))
